//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* file_out.cpp                                                           *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-, this file 2009-               *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h>
#include <map>
#include <iostream>

#include "mystd.h"
#include "olm.h"
#include "asi.h"
#include "file_out.h"

using namespace std;

namespace NFileOut {
    void Save_y08( FILE *file, const CEle* e );
    void Save_a01( FILE *file, const CEle* e );
    void Save_csv( FILE *file, const CEle* e );
    void Save_tab( FILE *file, const CEle* e );
}


//! Save files. Loop over file selection, and switch to format-sepcific calls.

void NFileOut::Save( string fmt )
{
    COlo *f;
    string outfnam;
    FILE *file;
    CEle *e;
    NOlm::IterateEle iter;
    while((e=iter())) {
        f = e->P();
        // query output file name
        outfnam = wask( "Save as (."+fmt+")", f->name);
        if ( outfnam=="" ) return;
        f->name = outfnam;
        outfnam += "."+fmt;
        // document file save
        f->lDoc.push_back( "fs " + outfnam + " # " +
                           mystd::myasctime(mystd::justnow(),0) );
        // prevent unintended overwriting
        if ( (file = fopen(outfnam.c_str(), "r")) ) {
            fclose(file);
            if (!bask("Overwrite ?")) return;
        }
        // save file
        if ( !(file = fopen(outfnam.c_str(), "w")) )
            throw "cannot write to file " + outfnam;
        if( fmt=="y08"){
            Save_y08( file, e );
        } else if( fmt=="a01"){
            Save_a01( file, e );
        } else if( fmt=="csv" ) {
            Save_csv( file, e );
        } else if( fmt=="tab" ) {
            Save_tab( file, e );
        } else
            throw string( "Unknown format" );
        fclose(file);
        f->as_on_disk = true;
    }
}


// Save file in format y08.

void NFileOut::Save_y08( FILE *file, const CEle* e )
{
    COlo* f; f = e->P();
    COld* d; d = e->D();
    COlc* c; c = e->C();
    string out;
    fprintf( file, "Meta:\n"
             "  format:  frida/y08 for yaml1\n"
             "  type:    %s\n",
             d ? "generic tabular data" : "frida2 curve" );

    fprintf( file, "History:\n" );
    for ( uint i=0; i<f->lDoc.size(); i++ )
        fprintf( file, "- %s\n", yaml(f->lDoc[i]).c_str() );
    fprintf( file, "- \%s\n", yaml( out ).c_str() );

    fprintf( file, "Coord:\n" );
    fprintf( file, "  x:\n    name: %s\n    unit: %s\n", 
             yaml(f->xco.name).c_str(), yaml(f->xco.unit).c_str() );
    fprintf( file, "  y:\n    name: %s\n    unit: %s\n", 
             yaml(f->yco.name).c_str(), yaml(f->yco.unit).c_str() );
    for ( uint i=0; i<f->ZCo.size(); i++ )
        fprintf( file, "  z%u:\n    name: %s\n    unit: %s\n", 
                 i, yaml(f->ZCo[i].name).c_str(),
                 yaml(f->ZCo[i].unit).c_str() );

    fprintf( file, "Param:\n" );
    for ( uint i=0; i<f->RPar.size(); i++ )
        fprintf( file,
                 "  -\n    name:  %s\n    unit:  %s\n    value: %20.15g\n",
                 yaml(f->RPar[i].Co.name).c_str(),
                 yaml(f->RPar[i].Co.unit).c_str(),
                 f->RPar[i].val );

    if( c ) {
        fprintf( file, "Formula: %s\n", c->expr.c_str() );
    }

    fprintf( file, "Tables: # %zu tables follow\n", f->nScan() );

    for( size_t j=0; j<f->nScan(); j++ ){
        fprintf( file, "  - # table %zu\n", j );
        if   ( d ) {
            CScan *s; s = &(d->VS[j]);
            for ( size_t i=0; i<f->nZ(); i++ ) 
                fprintf( file, "    z%zu: %18.10g\n", i, s->z[i] );
            fprintf( file, "    xy: |2 # %zu entries\n", s->size() );
            for( size_t i=0; i<s->size(); i++ ) 
                fprintf( file, "      %20.15g %20.15g\n", s->x[i], s->y[i] );
        } else {
            CCurve *s; s = &(c->VC[j]);
            for ( size_t i=0; i<f->nZ(); i++ ) 
                fprintf( file, "    z%zu: %18.10g\n", i, s->z[i] );
            for ( size_t i=0; i<c->nPar(); i++ ) 
                fprintf( file, "    p%zu: %18.10g\n", i, s->P[i] );
        }
    }
}


//! Save file in format ASCII-01.

void NFileOut::Save_a01( FILE *file, const CEle* e )
{
    COlo* f; f = e->P();
    COld* d; d = e->D();
    fprintf(file, "ASCII-01 JWu\n");
    if( d )
        fprintf(file, "typ D\n");
    else
        throw string( "ASCII-01 save only for data files" );
    fprintf( file, "fil %s\n", f->name.c_str() );
    for ( uint i=0; i<f->lDoc.size(); i++ )
        fprintf( file, "doc %s\n", f->lDoc[i].c_str() );

    fprintf( file, "x   %s\n", f->xco.to_a01().c_str() );
    fprintf( file, "y   %s\n", f->yco.to_a01().c_str() );
    for ( uint i=0; i<f->ZCo.size(); i++ )
        fprintf( file, "z%-2u %s\n", i, f->ZCo[i].to_a01().c_str() );
    for ( uint i=0; i<f->RPar.size(); i++ )
        fprintf( file, "rpa %20.15g %s\n",
                 f->RPar[i].val, f->RPar[i].Co.to_a01().c_str() );
    
    fprintf( file, "#j  %zu\n", f->nScan() );
    for( size_t j=0; j<f->nScan(); j++ ){
        CScan *s; s = &(d->VS[j]);
        fprintf( file, "scn %zu\n", j );
        fprintf( file, "Z   %zu\n", d->nZ() );
	for ( uint i=0; i<d->nZ(); i++ ) 
            fprintf( file, "%18.10g\n", s->z[i] );
	fprintf( file, "XY  %zu\n", s->size() );
        for( uint i=0; i<s->size(); i++ ) 
            fprintf( file, "%20.15g %20.15g\n", s->x[i], s->y[i] );
    }
    fprintf(file, "eof\n");
    f->as_on_disk = true;
}


//! Write data to table.

void NFileOut::Save_csv( FILE *file, const CEle* e )
{
    COlo* f; f = e->P();
    COld* d; d = e->D();
    if( !d )
        throw string( "csv save only for data files" );
    for( size_t j=0; j<f->nScan(); j++ ){
        for( size_t i=0; i<d->VS[j].size(); ++i ){
            fprintf( file, "%g\t", d->VS[j].y[i] );
        }
        fprintf( file, "\n" );
    } 
}


//! Write data to table.

void NFileOut::Save_tab( FILE *file, const CEle* e )
{
    COlo* f; f = e->P();
    COld* d; d = e->D();
    if( !d )
        throw string( "tab save only for data files" );
    for( size_t j=0; j<f->nScan(); j++ ){
        for( size_t iz=0; iz<f->ZCo.size(); ++iz ){
            if( iz>0 )  fprintf( file, "\t" );
            fprintf( file, "%g", d->VS[j].z[iz] );
        }
        fprintf( file, "\n" );
        for( size_t i=0; i<d->VS[j].size(); ++i ){
            fprintf( file, "%g\t%g\n", d->VS[j].x[i], d->VS[j].y[i] );
        }
        fprintf( file, "\n" );
    } 
}
