#include "boost/shared_ptr.hpp"

#include "list.h"
#include "coord.h"
#include "scan.h"
#include "curve.h"
#include "func.h"

//! Online object: virtual base class for COld, COlc.

class COlo {
 public:
    // data records:
    string name;
    vector<string> lDoc;
    CCoord xco, yco;
    vector<CCoord> ZCo;
    vector<CParam> RPar;
    // online state records:
    bool as_on_disk;
    CList ProtectedScans;

    virtual inline ~COlo() {}; // magic: makes dynamic_cast possible

    virtual void clear_V() = 0;
    virtual void erase( uint j ) = 0;
    virtual size_t nScan() const = 0;
    virtual double z( uint j, uint iz ) const = 0;
    virtual void erase_z( uint iz ) = 0;

    inline uint nZ() const { return ZCo.size(); };
};

//! Online data file.

class COld : public COlo {
 public:
    vector<CScan> VS;

    inline COld() {};
    COld( class COlc const* c );

    inline void clear_V() { VS.clear(); };
    inline void erase( uint j ) { VS.erase( VS.begin()+j ); };
    inline size_t nScan() const { return VS.size(); };
    inline double z( uint j, uint iz ) const { return VS[j].z[iz]; };
    inline void erase_z( uint iz ){
        for( uint j=0; j<nScan(); ++j )
            VS[j].z.erase( VS[j].z.begin()+iz );
    };

    uint nPts(uint j) const; 
    uint nPts() const; // 0 unless all spectra j have same nPts(j)
};

//! Online curve file.

class COlc : public COlo { 
 public:
    // basic records (offline and online):
    string expr;
    vector<CCoord> PCo;
    vector<CCurve> VC;

    // online state records:
    boost::shared_ptr<class CTree> T;
    vector<CList>  Fixed;
    bool           weight_y_log;
    uint           kd; // internal number of data file to be fitted
    uint           kconv;
    uint           cv_intmod;
    double         cv_epsabs;
    double         cv_epsrel;

    inline COlc() : weight_y_log(false), kconv(-1) { ; };

    inline void clear_V() { VC.clear(); };
    inline void erase( uint j ) { VC.erase( VC.begin()+j ); };
    inline double z( uint j, uint iz ) const { return VC[j].z[iz]; };
    inline size_t nScan() const { return VC.size(); };
    inline void erase_z( uint iz ){
        for( uint j=0; j<nScan(); ++j )
            VC[j].z.erase( VC[j].z.begin()+iz );
    };

    inline uint nPar() const { return PCo.size(); };
    vector<string> pInfo() const;
    string pInfoCat() const;
};

//! Element of online memory: pointer to COlo.

class CEle {
    enum eleTyp { DATA, CURVE }; 
    eleTyp Typ;
    void* Ptr;

 public:
    inline CEle() {};
    inline CEle( COld* pin ) {
        Typ = DATA;
        Ptr = (void*) pin;
    };
    inline CEle( COlc* pin ) {
        Typ = CURVE;
        Ptr = (void*) pin;
    };
    // inline ~CEle(void) {  };
	
    inline bool isD() const { return Typ==DATA; };

    COlo* copy() const {
        COlo* ret;
        if( Typ==DATA ){
            ret = (COlo*) new COld();
            *((COld*)ret) = *((COld*) Ptr);
        } else {
            ret = (COlo*) new COlc();
            *((COlc*)ret) = *((COlc*) Ptr);
        }
        return ret;
    }

    COld* D() const {
        if (Typ==DATA)  return (COld*) Ptr;
        else            return 0;
    };
    COlc* C() const {
        if (Typ==CURVE) return (COlc*) Ptr;
        else            return 0;
    };
    COlo* P() const { return (COlo*) Ptr; };

    void predelete() { 
        if      (Typ==DATA)  delete (COld*) Ptr;
        else if (Typ==CURVE) delete (COlc*) Ptr;
        else throw string( "FATAL ERROR in CEle:predelete" );
    };

    uint nScan() const {
        if      (Typ==DATA)  return ((COld*) Ptr)->nScan();
        else if (Typ==CURVE) return ((COlc*) Ptr)->nScan();
        else throw string( "FATAL ERROR in CEle:nScan" );
    };
    uint nPts() const {
        if      (Typ==DATA)  return ((COld*) Ptr)->nPts();
        else if (Typ==CURVE) return ((COlc*) Ptr)->PCo.size();
        else throw string( "FATAL ERROR in CEle:nPts" );
    };
    vector<double> *Z(uint j) const {
        if      (Typ==DATA)  return &(((COld*) Ptr)->VS[j].z);
        else if (Typ==CURVE) return &(((COlc*) Ptr)->VC[j].z);
        else throw string( "FATAL ERROR in CEle:Z()" );
    };
};
		
//! The online memory.

namespace NOlm {
    // the data:
    extern vector<CEle> MOM; // the mother of all on-line data

    // the current selection:
    extern CList FSel;
    extern CList FSelOld;
    extern bool overwrite;

    // modify data:
    void SelNew(void);
    void SelPop(void);
    CEle* OloPushBack( CEle *e );
    CEle* OloAdd( const class COlo *olo );
    void OlfDel();
    void OlfExchange();
    void OlfCopy();
    void OlfLoop(void (*exec) (int fno, COld *fin, COld *fout));

    // scan selection:
    void JSelAsk( string quest, string *jSel, CList *JSel );

    class Iterate { // iterator over FSel (virtual base)
    protected:
        uint iV;
        CList Sel;
    public:
        inline Iterate(void) { Sel = FSel; Reset(); };
        inline void Reset(void) { iV = 0; };
        inline void All(void) { Sel = CList(0,MOM.size()-1); };
        inline uint SelNo(void) const { return Sel.V[iV-1]; }
        inline uint index(void) const { return iV-1; }
        inline uint size(void) const { return Sel.V.size(); }
    };
    class IterateEle : public Iterate { // all types
    public:
        inline CEle* operator()(void) {
            if (iV>=Sel.size()) return 0;
            return &(MOM[Sel.V[iV++]]);
        };
    };
    class IterateD : public Iterate { // must be data file
        COld *e;
    public:
        inline COld* operator()(void) {
            do {
                if (iV>=Sel.size()) return 0;
            } while (! (e = MOM[Sel.V[iV++]].D()));
            return e;
        };
    };
    class IterateC : public Iterate { // must be curve file
        COlc *e;
    public:
        inline COlc* operator()(void) {
            do {
                if (iV>=Sel.size()) return 0;
            } while (! (e = MOM[Sel.V[iV++]].C()));
            return e;
        };
    };
};

//! Chunks of frequently used code.

#define OUT_TO_IN( overwrite, fin, ftmp, fout ) \
    if (overwrite) {                            \
        fout = fin;                             \
    } else {                                    \
        ftmp = *fin;                            \
        fout = &ftmp;                           \
    }
